"""
BaseTest subclasses unittest.TestCase to provide a base class shared by
testclasses across the 'tests' package. It implements module unloading in
the 'tearDownClass' method and adds 'assertIsSubclass' (and negation).
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from unittest import TestCase
from random import randint, random
from typing import TYPE_CHECKING

from ..desc import Field
from ..utilities import maybe

if TYPE_CHECKING:  # pragma: no cover
  from typing import TypeAlias, Iterator

  IntSample: TypeAlias = Iterator[tuple[int, ...]]
  TypicalExceptions: TypeAlias = Iterator[type[Exception]]
  ExcType: TypeAlias = type[BaseException]

#  So Python 3.14 saw 'assertIsSubclass' and 'assertNotIsSubclass' added to
#  'unittest.TestCase', but since we support back to 3.7, but also up to
#  3.14, we need the following 'try-except-else':

try:
  _ = TestCase.assertIsSubclass
except AttributeError:  # pragma: no cover
  class _Temp(TestCase):
    def assertIsSubclass(self, cls: type, base: type, msg=None) -> None:
      """Assert that 'subClass' is a subclass of 'superClass'."""
      self.assertTrue(issubclass(cls, base))

    def assertNotIsSubclass(self, cls: type, base: type, msg=None) -> None:
      """Assert that 'subClass' is not a subclass of 'superClass'."""
      self.assertFalse(issubclass(cls, base))
else:  # version >= 3.14
  _Temp = TestCase


class BaseTest(_Temp):
  """
  BaseTest provides a base class shared by the testing classes in the
  tests package. It implements module unloading in the 'tearDownClass'
  method and adds 'assertIsNotInstance' and 'assertIsNotSubclass'.
  """

  attrErrTrace = Field()
  exceptions = Field()

  @attrErrTrace.GET
  def _getAttributeErrorTrace(self, ) -> str:
    return """object has no attribute"""

  @exceptions.GET
  def _getTypicalExceptions(self, ) -> TypicalExceptions:
    yield from (
      ValueError,
      KeyError,
      IndexError,
      RuntimeError,
      OSError,
      PermissionError,
      TypeError,
      FileExistsError,
      FileNotFoundError,
      )

  @staticmethod
  def generateRandomIntegers(*args) -> IntSample:
    """
    Generates a list of tuples containing random integers.
    Arguments:
      args = (n, N, a, b)
        n: Number of tuples to generate. Default is 1.
        N: Number of integers in each tuple. Default is 1.
        a: Minimum integer value (inclusive). Default is 0.
        b: Maximum integer value (exclusive). Default is 256.
      if a > b, then a and b are swapped.
    Returns:
      A list of 'N' tuples, each containing 'n' random integers in
      range [a, b).
    """
    n, N, a, b, *_ = [*args, None, None, None, None]
    n, N, a, b = maybe(n, 1), maybe(N, 1), maybe(a, 0), maybe(b, 256),
    a, b = (a, b) if a < b else (b, a)
    for _ in range(N):
      yield tuple(randint(a, b - 1) for _ in range(n))

  @staticmethod
  def randFloat(*args) -> float:
    a, b, *_ = (*args, 0.0, 1.0)
    minVal, maxVal = min(a, b), max(a, b)
    return random() * (maxVal - minVal) + minVal

  @classmethod
  def randFloats(cls, N: int = None, *args) -> Iterator[float]:
    for _ in range(maybe(N, 1)):
      yield cls.randFloat(*args)

  @classmethod
  def randFloatTuple(cls, N: int = None, *args) -> tuple[float, ...]:
    return (*cls.randFloats(N, *args),)

  @classmethod
  def randFloatTuples(cls, *args) -> Iterator[tuple[float, ...]]:
    n, N, a, b, *_ = (*args, None, None, None, None)
    n, N = maybe(n, 1), maybe(N, 1)
    a, b = maybe(a, 0.0), maybe(b, 1.0)
    minVal, maxVal = min(a, b), max(a, b)
    for _ in range(n):
      yield cls.randFloatTuple(N, minVal, maxVal)

  @classmethod
  def tearDownClass(cls) -> None:
    """Remove the test class from sys.modules and run the garbage
    collector."""
    import sys
    import gc
    sys.modules.pop(cls.__module__, None)
    gc.collect()

  assertIsNotSubclass = _Temp.assertNotIsSubclass
  assertIsNotInstance = _Temp.assertNotIsInstance



"""
KeeBox subclasses AttriBox and provides special support for the
enumerations. This solves a general problem with 'AttriBox' objects using
enumerations as their field types. When doing so, the positional arguments
in the parentheses, *must* be a member of the enumeration. First of all,
this requires duplication of the enumeration class. Secondly, it fails to
provide the intended flexibility of passing constructor arguments for
deferred instantiation. Because enumerations instantiate *during* class
creation, there is no deferred instantiation. Finally, when an descriptor
is retrieved from an instance, it is mostly for the purpose of retrieving
the value of the descriptor.

The KeeBox makes the following attempts at resolving arguments to a member
of the enumeration:
1: Single 'str' object received
1a: The string matches the 'name' of a member of the enumeration.
1b: The string matches the 'value' of a member of the enumeration.
2: Single 'int' object received
2a: The integer matches the 'index' of a member of the enumeration.
2b: The integer matches the 'value' of a member of the enumeration.
3: Single argument received of the 'valueType' type of the enumeration.
3a: The argument matches the 'value' of a member of the enumeration.
4: Any number of arguments received
4a: If the enumeration class is a 'KeeFlags' class instead. Then for each
argument, it attempts to resolve it to a flag member. If all succeed,
it returns the member having those flags high.

The above process applies to both instantiation and setting. Please note
that 'KeeBox' will never attempt to instantiate the 'valueType' of the
enumeration.

Examples:

  #  KeeNum classes:
  class KeyboardNum(KeeNum):
    A = Kee[str]('key A')
    B = Kee[str]('key B')
    ...

  class KeyModFlags(KeeFlags):
    SHIFT = KeeFlag(0)
    CTRL = KeeFlag(1)
    ALT = KeeFlag(2)
    META = KeeFlag(3)

  #  Illustrative classes using the above KeeNum classes:
  class SelectAll:
    key = KeeBox[KeyboardNum]('A')
    mod = KeeBox[KeyModFlags]('ctrl')

  class Settings:  # PyCharm: CTRL+SHIFT+S
    key = KeeBox[KeyboardNum]('S')
    mod = KeeBox[KeyModFlags]('ctrl', 'shift')
"""
#  AGPL-3.0 license
#  Copyright (c) 2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import KeeMeta, KeeFlagsMeta
from ..desc import AttriBox
from ..waitaminute import TypeException
from ..waitaminute.keenum import KeeBoxException, \
  KeeBoxValueError, KeeBoxTypeError

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Self


class KeeBox(AttriBox):
  """
  KeeBox subclasses AttriBox and provides special support for the
  enumerations.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Class Variables

  #  Fallback Variables

  #  Private Variables

  #  Public Variables

  #  Virtual Variables

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __get__(self, instance: Any, owner: type, **kwargs) -> Any:
    if instance is None:
      return self
    pvtName = self.getPrivateName()
    try:
      value = getattr(instance, pvtName)
    except AttributeError as attributeError:
      if kwargs.get('_recursion', False):
        raise RecursionError from attributeError
      try:
        fieldObject = self._resolve()
      except Exception as exception:
        raise exception from attributeError
      else:
        setattr(instance, pvtName, fieldObject)
        return self.__get__(instance, owner, _recursion=True)
    else:
      return value

  def __set__(self, instance: Any, value: Any, **kwargs) -> None:
    pvtName = self.getPrivateName()
    fieldNum = self.fieldType
    if isinstance(value, fieldNum):
      return setattr(instance, pvtName, value)
    if kwargs.get('_recursion', False):
      raise RecursionError
    return self.__set__(instance, self._resolve(value, ), _recursion=True)

  def _resolve(self, *args, **kwargs) -> Any:
    fieldNum = self.fieldType
    if isinstance(fieldNum, KeeMeta):
      return self._resolveNum(*args, )
    elif isinstance(fieldNum, KeeFlagsMeta):
      return self._resolveFlags()
    else:
      raise KeeBoxException(self, self.getPosArgs())

  def _resolveNum(self, *args, ) -> Any:
    args = args or self.getPosArgs()
    kwargs = self.getKeyArgs()
    fieldNum = self.fieldType
    valueType = fieldNum.valueType
    if len(args) == 1:
      if isinstance(args[0], str):
        for member in fieldNum:
          if args[0] == member.name:
            return member
        for member in fieldNum:  # Case-insensitive match
          if str.lower(args[0]) == str.lower(str(member.name)):
            return member
      if isinstance(args[0], int):
        if args[0] < len(fieldNum):
          return fieldNum[args[0]]
        for member in fieldNum:
          if args[0] == member.value:
            return member
      elif isinstance(args[0], valueType):
        for member in fieldNum:
          if args[0] == member.value:
            return member
    try:
      tempObject = valueType(*args, **kwargs)
    except Exception as exception:
      raise KeeBoxTypeError(self, *args) from exception
    else:
      for member in fieldNum:
        if tempObject == member.value:
          return member
      else:
        raise KeeBoxValueError(self, fieldNum, tempObject)

  def _resolveFlags(self, ) -> Any:
    highs, args, kwargs = [], self.getPosArgs(), self.getKeyArgs()
    for arg in args:
      highs.append(self._resolveNum(arg, ))
    names = frozenset((*(h.name for h in highs),), )
    return self.fieldType.memberDict[names]

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def __class_getitem__(cls, fieldType: Any) -> Self:
    """
    Before calling 'AttriBox.__class_getitem__', it ensures that the item
    given is a class derived from either 'KeeMeta' or 'KeeFlagsMeta'. If
    not, it will raise 'TypeException'. By restricting 'KeeBox' this way,
    it greatly simplifies the implementation. In those cases, the existing
    'AttriBox' provides the necessary implementation. Please note however,
    that 'AttriBox' provides no such check and will accept every 'type',
    even those derived from 'KeeMeta'  or 'KeeFlagsMeta'.
    """
    if isinstance(fieldType, KeeMeta) or isinstance(fieldType, KeeFlagsMeta):
      return super().__class_getitem__(fieldType)
    raise TypeException('fieldType', fieldType, KeeMeta, KeeFlagsMeta)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #



"""
KeeBoxException is a custom exception class raised to indicate that a
'KeeBox' object failed to resolve to the underlying member.
"""
#  AGPL-3.0 license
#  Copyright (c) 2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class KeeBoxException(Exception):
  """
  KeeBoxException is a custom exception class raised to indicate that a
  'KeeBox' object failed to resolve to the underlying member.
  """

  __slots__ = ('box', 'args')

  def __init__(self, box: Any, args: Any) -> None:
    self.box, self.args = box, args
    Exception.__init__(self, )

  def __str__(self) -> str:
    infoSpec = """KeeBox object: '%s' failed to resolve with arguments: 
    <br><tab>%s<br>"""
    boxStr = str(self.box)
    argSpec = """<%s: %s>"""
    argTypes = (*(type(arg).__name__ for arg in self.args),)
    argStrs = (*(repr(arg) for arg in self.args),)
    argInfos = (*(argSpec % (t, s) for s, t in zip(argStrs, argTypes)),)
    argInfo = '<br><tab>'.join(argInfos)
    info = infoSpec % (boxStr, argInfo)
    from ...utilities import textFmt
    return textFmt(info, )

  __repr__ = __str__



"""
The 'worktoy.work_test' package provides a subclass of 'unittest.TestCase'.
"""
#  AGPL-3.0 license
#  Copyright (c) 2026 Asger Jon Vistisen
from __future__ import annotations

from ._base_test import BaseTest

__all__ = [
  'BaseTest',
  ]



"""DispatchException provides a custom exception raised when an instance
of OverloadDispatcher fails to resolve the correct function from the
given arguments. Because the overload protocol relies on type matching,
this exception subclasses TypeError such that it can be caught by external
error handlers. """
#  AGPL-3.0 license
#  Copyright (c) 2024-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, TypeAlias, Union

  Args: TypeAlias = tuple[Any, ...]
  Excs: TypeAlias = tuple[Exception, ...]

  from worktoy.dispatch import Dispatcher
  from worktoy.dispatch import overload

  Overloaded: TypeAlias = Union[Dispatcher, overload]
  #  At runtime, the 'overload' objects will be replaced with 'Dispatcher'
  #  objects, but that is not visible to type checkers.


class DispatchException(TypeError):
  """
  DispatchException provides a custom exception raised to indicate that a
  'Dispatcher' object failed to resolve given arguments to a matching
  function object.

  Arguments:
    dispatch: The 'Dispatcher' object that failed to dispatch the given
    arguments. (When manually raising this exception, the static type
    checker will recognize the relevant objects as 'overload' objects,
    but at runtime these will be replaced with 'Dispatcher' objects.
    Hinting both types prevents type checkers from indicating type
    mismatch.)
    args: The arguments that could not be dispatched
  """

  __slots__ = ('dispatch', 'args')

  def __init__(self, dispatch: Overloaded, args: Args, ) -> None:
    self.dispatch = dispatch
    self.args = args
    TypeError.__init__(self, )

  def __str__(self) -> str:
    """
    Return a string representation of the DispatchException.
    """
    infoSpec = """Dispatcher object: '%s' failed to dispatch arguments: 
    <br><tab>%s<br><tab>matching type signature: '%s'<br>"""
    dispStr = str(self.dispatch)
    argsStr = ', '.join(str(arg) for arg in self.args)
    from ...dispatch import TypeSig
    typeStr = str(TypeSig.fromArgs(*self.args))
    info = infoSpec % (dispStr, argsStr, typeStr)
    return textFmt(info, )

  __repr__ = __str__



"""
KeeBoxValueError is a custom exception class raised to indicate that a
value specified in the positional arguments of a 'KeeBox' object does
match a valid object of the value type of the given enumeration type.
"""
#  AGPL-3.0 license
#  Copyright (c) 2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  pass


class KeeBoxValueError(ValueError):
  """
  KeeBoxValueError is a custom exception class raised to indicate that a
  value specified in the positional arguments of a 'KeeBox' object does
  match a valid object of the value type of the given enumeration type.
  """

  __slots__ = ('desc', 'num', 'value')

  def __init__(self, *args) -> None:
    self.desc, self.num, self.value, *_ = (*args, None, None, None)
    ValueError.__init__(self, )

  def __str__(self, ) -> str:
    infoSpec = """The '%s' descriptor with field enumeration: '%s', 
    having value type: '%s', received value of correct type: '%s', 
    which does not match the 'value' of any member of the '%s' 
    enumeration!"""
    descSpec = """%s.%s: %s"""
    ownerName = getattr(self.desc, '__field_owner__', ).__name__
    fieldName = getattr(self.desc, '__field_name__', )
    fieldType = str(getattr(self.desc, '__field_type__', ))
    desc = descSpec % (ownerName, fieldName, fieldType)
    num = str(self.num)
    valueType = str(type(self.value))
    info = infoSpec % (desc, num, valueType, repr(self.value), fieldType)
    from ...utilities import textFmt
    return textFmt(info, )

  __repr__ = __str__



"""
KeeMeta provides the metaclass for the 'worktoy.keenum' module.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ._kee_desc import Base, MRO, Members, IsRoot
from ..core import Object
from ..desc import Field
from ..mcls import BaseMeta
from ..waitaminute import TypeException, attributeErrorFactory
from ..waitaminute.keenum import KeeNameError, KeeIndexError
from ..waitaminute.keenum import KeeMemberError, KeeValueError
from . import KeeSpace as KSpace, Kee

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, TypeAlias, Self, Iterator

  from . import KeeNum

  Bases: TypeAlias = tuple[type, ...]


class KeeMeta(BaseMeta):
  """
  KeeMeta provides the metaclass for the 'worktoy.keenum' module.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __allow_instantiation__ = False
  __num_members__ = None
  __is_root__ = None

  #  Public Variables
  isRoot = IsRoot()
  base = Base()
  mroNum = MRO()
  members = Members()
  valueType = Field()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @valueType.GET
  def _getValueType(cls, ) -> type:
    """Gets the value type of the enumeration. """
    type_ = None
    for member in cls.__num_members__:
      valType = type(member.value)
      if type_ is None:
        type_ = valType
    return type_

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def __prepare__(mcls, name: str, bases: Bases, **kw: Any) -> KSpace:
    """Prepares the namespace for the class."""
    bases = (*[b for b in bases if b.__name__ != '_InitSub'],)
    return KSpace(mcls, name, bases, **kw)

  def __new__(mcls, name: str, bases: Bases, space: KSpace, **kw) -> Self:
    """
    Creates a new instance of the class.
    This method is called when the class is created.
    """
    cls = super().__new__(mcls, name, bases, space, **kw)
    cls.__num_members__ = []
    if not [b for b in bases if b is not Object]:
      return cls
    cls.__allow_instantiation__ = True

    for i, (key, kee) in enumerate(space.__enumeration_members__.items()):
      self = None
      for base in bases:
        self = getattr(base, key, None)
      if self is None:
        self = cls(kee, )
      setattr(cls, key, self)
      cls.__num_members__.append(self)
    cls.__allow_instantiation__ = False
    return cls

  def __call__(cls, *args, **kwargs) -> Any:
    """
    Prevents instantiation of the class.
    This method is called when the class is instantiated.
    """
    if cls.__allow_instantiation__:
      return super().__call__(*args, **kwargs)
    return cls._resolveMember(args[0])

  def __getitem__(cls, identifier: Any) -> Any:
    """
    Gets a member of the enumeration by index or key.
    This method is used to get a member of the enumeration by index or key.
    """
    return cls._resolveMember(identifier)

  def __getattr__(cls, name: str) -> Any:
    """
    Gets a member of the enumeration by name.
    This method is used to get a member of the enumeration by name.
    """
    try:
      self = cls._resolveKey(name)
    except KeeNameError as keeNameError:
      error = attributeErrorFactory(cls, name)
      raise error from keeNameError
    else:
      return self

  def __iter__(cls, ) -> Iterator[Self]:
    """
    Iterates over the members of the enumeration.
    This method is used to iterate over the members of the enumeration.
    """
    yield from cls.members

  def __len__(cls, ) -> int:
    """
    Returns the number of members in the enumeration.
    This method is used to get the number of members in the enumeration.
    """
    return sum(1 for _ in cls)

  def __contains__(cls, identifier: Any) -> bool:
    """
    Checks if the enumeration contains a member by index or key.
    This method is used to check if the enumeration contains a member.
    """
    try:
      _ = cls._resolveMember(identifier)
    except (IndexError, KeyError):
      return False
    else:
      return True

  def __instancecheck__(cls, instance: Any) -> bool:
    """
    Checks if the instance is a member of the enumeration.
    This method is used to check if the instance is a member of the
    enumeration.
    """
    for member in cls:
      if member == instance:
        return True
    if issubclass(type(instance), cls):
      return True
    return False

  def __subclasscheck__(cls, subclass: type) -> bool:
    """
    Checks if the subclass is a subclass of the enumeration.
    This method is used to check if the subclass is a subclass of the
    enumeration.
    """
    try:
      _ = issubclass(subclass, object)
    except Exception as exception:
      raise exception
    else:
      for item in subclass.__mro__:
        if item is cls:
          return True
      return False

  def __str__(cls) -> str:
    """
    Returns a string representation of the enumeration.
    This method is used to get a string representation of the enumeration.
    """
    infoSpec = """<KeeNum '%s': %d members>"""
    clsName = cls.__name__
    n = len(cls.__num_members__)
    info = infoSpec % (clsName, n)
    return info

  __repr__ = __str__

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _resolveMember(cls, identifier: Any) -> Any:
    """
    Resolves a member of the enumeration.
    This method is used to resolve a member of the enumeration.
    """
    if isinstance(identifier, int):
      return cls._resolveIndex(identifier)
    if isinstance(identifier, str):
      return cls._resolveKey(identifier)
    if type(type(identifier)) is type(cls):
      for member in cls:
        if member is identifier:
          return member
      raise KeeMemberError(cls, identifier)
    try:
      member = cls.fromValue(identifier)
    except KeeValueError:
      from . import KeeNum
      raise TypeException('identifier', identifier, int, str, KeeNum)
    else:
      return member

  def _resolveIndex(cls, index: int) -> Kee:
    """
    Resolves a member of the enumeration by index.
    This method is used to resolve a member of the enumeration by index.
    """
    if index < 0:
      return cls[len(cls) + index]
    if index < len(cls):
      return cls.__num_members__[index]
    raise KeeIndexError(cls, index)

  def _resolveKey(cls, key: str) -> Kee:
    """
    Resolves a member of the enumeration by key.
    This method is used to resolve a member of the enumeration by key.
    """
    for member in cls.__num_members__:
      if member.name.lower() == key.lower():
        return member
    raise KeeNameError(cls, key)

  def fromValue(cls, value: Any) -> KeeNum:
    """
    Resolves a member of the enumeration by value.
    This method is used to resolve a member of the enumeration by value.
    """
    for member in cls.__num_members__:
      if member.value == value:
        return member
    raise KeeValueError(cls, value)



"""
The 'worktoy.examples' module provides example usages of the 'worktoy'
library.

Usage:
In a terminal:
python -m worktoy.examples <name_of_example>

For example,
python -m worktoy.examples complex_number

The available examples are:

"""
#  AGPL-3.0 license
#  Copyright (c) 2026 Asger Jon Vistisen
from __future__ import annotations

from ._complex_number import ComplexNumber

__all__ = [
  'ComplexNumber',
  ]



"""
AttriBox implements a lazily instantiated and strongly typed descriptor
class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import Field
from ..core import Object
from ..core.sentinels import DELETED
from ..utilities import typeCast
from ..waitaminute import TypeException
from ..waitaminute.dispatch import TypeCastException

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Self


class AttriBox(Object):
  """
  AttriBox implements a lazily instantiated and strongly typed descriptor
  class.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __field_type__ = None  # The type of the objects stored in the box.

  #  Public Variables
  fieldType = Field()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @fieldType.GET
  def getFieldType(self) -> type:
    return self.__field_type__

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _resolve(self, *args, **kwargs) -> Self:
    """
    Creates a new instance of the field type from the given arguments.
    Please note, that this method does *not* retrieve arguments from
    'self', but requires them to be passed in as arguments. This is because
    this method is used both when setting and getting.

    When '__instance_get__' is unable to retrieve a value from a given
    instance. The 'args' and 'kwargs' passed to the constructor of the
    'AttriBox' are retrieved from the 'getPosArgs' and 'getKeyArgs'
    methods, and passed to this method.

    The '__instance_set__' method will always receive one object. However,
    if this object is a 'tuple', the intention may be that these should be
    understood as starred arguments. For example:

    class Foo:
      bar = AttriBox[complex](0)

      def __getitem__(self, index: Any) -> Any: ...

    foo = Foo()
    foo.bar = 69, 420

    foo.bar = 69, 420
    #  The descriptor protocol processes the above assignment as:
    AttriBox.__set__(Foo.bar, foo, (69, 420))
    #  Similarly to how:
    foo[69, 420]
    #  is processed as:
    Foo.__getitem__(foo, (69, 420))
    In the above examples, it is *not* possible to pass keyword arguments.
    This is the only difference between function calls and the above. For
    this reason, a received 'tuple' object implies a sequence of
    positional arguments. When passing such on to the field type
    constructor, it is reasonable to infer that a star (*) were
    intended. With a few exceptions, described below. Please note,
    that 'args' in any 'function' object with starred arguments,
    is a 'tuple' of the passed arguments. The tuple spoken off here,
    is when the 'args' object consists only of one 'tuple' object.

    The exceptions referenced above all have to do with builtin collection
    types. These demand receiving a tuple as a single argument, rather
    than an arbitrary number of arguments. For example, the following is
    not valid:
    sus = list(1, 2, 3)
    but the following is:
    meh = list((1, 2, 3))  # no star
    The above is true for: tuple, list, set and frozenset. For dict,
    each element of the tuple must itself be a tuple of length 2, with the
    first element being hashable. For 'set' and 'frozenset', every element
    must be hashable. The final exception is that in the presence of
    keyword arguments, the 'tuple' is not unpacked.
    """
    fieldType = self.getFieldType()
    try:
      if fieldType in (list, set, frozenset, dict, tuple):
        fieldObject = fieldType(args)
      else:
        fieldObject = fieldType(*args, **kwargs)
    except (TypeError, ValueError) as exception:
      name = 'value'
      raise TypeException(name, args[0], fieldType) from exception

    try:
      setattr(fieldObject, '__field_name__', self.getFieldName())
      setattr(fieldObject, '__field_owner__', self.getFieldOwner())
      setattr(fieldObject, '__field_box__', self)
    except AttributeError:
      pass
    return fieldObject

  def __instance_get__(self, instance: Any, owner: type, **kwargs) -> Any:
    """
    Returns the value of the field for the given instance. If the value is
    not set, it initializes it with a new instance of the field type.
    """
    pvtName = self.getPrivateName()
    try:
      value = getattr(instance, pvtName)
    except AttributeError as attributeError:
      if kwargs.get('_recursion', False):
        raise RecursionError from attributeError
      args = self.getPosArgs()
      kwargs = self.getKeyArgs()
      fieldObject = self._resolve(*args, **kwargs)
      setattr(instance, pvtName, fieldObject)
      return self.__instance_get__(instance, owner, _recursion=True)
    else:
      return value

  def __instance_set__(self, instance: Any, value: Any, **kwargs) -> None:
    """
    Sets the value of the field for the given instance. If the value is
    not set, it initializes it with a new instance of the field type.
    """
    fieldType = self.getFieldType()
    pvtName = self.getPrivateName()
    if isinstance(value, fieldType) or kwargs.get('_root', False):
      return setattr(instance, pvtName, value)
    #  Catch recursion
    if kwargs.get('_recursion', False):
      raise RecursionError
    #  Attempt to 'typeCast' without instantiation
    try:
      cast = typeCast(fieldType, value, allowInstantiation=False)
    except TypeCastException as typeCastException:
      if isinstance(value, tuple):
        args = (*(self.filterSentinels(arg) for arg in value),)
      else:
        args = (self.filterSentinels(value),)
      try:
        fieldObject = self._resolve(*args, **kwargs)
      except Exception as exception:
        raise exception from typeCastException
      else:
        return self.__instance_set__(instance, fieldObject, _recursion=True)
    else:
      return self.__instance_set__(instance, cast, _recursion=True)

  def __instance_delete__(
      self,
      instance: Any,
      old: Any = None,
      **kwargs,
      ) -> None:
    """
    Deletes the value of the field for the given instance. If the value is
    not set, it does nothing.
    """
    pvtName = self.getPrivateName()
    setattr(instance, pvtName, DELETED)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def __class_getitem__(cls, fieldType: type) -> Self:
    """
    Allows the AttriBox to be used as a generic type with a specified
    field type.
    """
    self = object.__new__(cls)
    self.__field_type__ = fieldType
    return self  # noqa

  def __call__(self, *args: Any, **kwargs: Any) -> Any:
    """
    Allows the AttriBox to be called like a function, returning a new
    instance of the field type.
    """
    Object.__init__(self, *args, **kwargs)
    return self

  def __delete__(self, instance: Any, **kwargs) -> None:
    try:
      Object.__delete__(self, instance, **kwargs)
    except AttributeError as attributeError:
      fieldName = self.getFieldName()
      raise AttributeError(fieldName) from attributeError



"""
ValidSlice is a special un-instantiable class that recognizes only valid
slice objects.
"""
#  AGPL-3.0 license
#  Copyright (c) 2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import textFmt

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Never


class _MetaSlice(type):

  def __instancecheck__(self, instance: Any) -> bool:
    if not isinstance(instance, slice):
      return False
    keys = ('start', 'stop', 'step',)
    for key in keys:
      val = getattr(instance, key)
      if val is None or isinstance(val, int):
        continue
      try:
        #  Check for '__index__' method
        valType = type(val)
        indexFunc = getattr(valType, '__index__', )

        #  Check  for '__index__' being callable
      except (AttributeError, TypeError):
        return False
      else:
        if not callable(indexFunc):
          return False
        continue
    return True

  def __call__(cls, *__, **_) -> Never:
    infoSpec = """%s cannot be instantiated. It is intended only for 
    validating 'slice' objects."""
    info = infoSpec % cls.__name__
    raise TypeError(textFmt(info))

  def __str__(cls, ) -> str:
    infoSpec = """<type '%s' | slice object validator>"""
    info = infoSpec % cls.__name__
    return textFmt(info)

  def __repr__(cls, ) -> str:
    infoSpec = """worktoy.utilities.%s"""
    info = infoSpec % cls.__name__
    return textFmt(info)


class ValidSlice(metaclass=_MetaSlice):
  """
  ValidSlice is a special un-instantiable class that recognizes only valid
  slice objects. Passing keyword arguments to the slice constructor does
  raise:
    'TypeError: slice() takes no keyword arguments'.
  Furthermore, passing no arguments at all likewise raises:
    'slice expected at least 1 argument, got 0'.
  Finally, passing arbitrary positional arguments raises:
    'slice indices must be integers or None or have an __index__ method'
  Or do they? Contrary to every other type and every reasonable expectation
  any *bonus pater familias* might have, slice objects waits until usage
  before raising the above exception. ValidSlice provides a special type,
  that recognizes slice objects as instances of itself, provided they are
  not malformed.

  For example:
  isinstance(slice('imma slice, trust me bro!'), ValidSlice) -> False
  isinstance(slice(0, 10, 'step'), ValidSlice) -> False
  isinstance(slice(1), ValidSlice) -> True
  isinstance(slice(None, None, None), ValidSlice) -> True
  """
  pass



"""
KeeBoxTypeError is a custom exception class raised to indicate that a
given 'KeeBox' descriptor could not resolve given arguments to member of
the value type of the field enumeration of the descriptor.
"""
#  AGPL-3.0 license
#  Copyright (c) 2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover

  from ...keenum import KeeBox


class KeeBoxTypeError(TypeError):
  """
  KeeBoxTypeError is a custom exception class raised to indicate that a
  given 'KeeBox' descriptor could not resolve given arguments to member of
  the value type of the field enumeration of the descriptor.
  """

  __slots__ = ('box', 'args')

  def __init__(self, box: KeeBox, *args) -> None:
    self.box, self.args = box, args
    TypeError.__init__(self, )

  def __str__(self, ) -> str:
    infoSpec = """The '%s' descriptor with field enumeration: '%s' having 
    value type: '%s' could not be instantiated with the given arguments: 
    <br><tab>%s<br>"""
    descSpec = """%s.%s: %s"""
    ownerName = getattr(self.box, '__field_owner__', ).__name__
    fieldName = getattr(self.box, '__field_name__', )
    fieldType = str(getattr(self.box, '__field_type__', ))
    desc = descSpec % (ownerName, fieldName, fieldType)
    num = str(self.box.fieldType)
    valueType = str(self.box.fieldType.valueType)
    argSpec = """<%s: %s>"""
    argTypes = (*(type(arg).__name__ for arg in self.args),)
    argStrs = (*(repr(arg) for arg in self.args),)
    argInfos = (*(argSpec % (t, s) for s, t in zip(argStrs, argTypes)),)
    argInfo = '<br><tab>'.join(argInfos)
    info = infoSpec % (desc, num, valueType, argInfo)
    from ...utilities import textFmt
    return textFmt(info, )

  __repr__ = __str__
